# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pytest


def test_clean_timeseries_no_tsfresh():
    sys.modules['tsfresh'] = None

    sys.modules.pop('pylife.stress.timesignal', None)

    import pylife.stress.timesignal as pts

    with pytest.raises(ImportError, match=r"pip install pylife\[tsfresh\]"):
        pts.clean_timeseries(None, None)
